%
%
%  Track Translate Function for Imaris 7.3.0
%
%  Copyright Bitplane AG 2011
%
%
%  Installation:
%
%  - Copy this file into the XTensions folder in the Imaris installation directory.
%  - You will find this function in the Image Processing menu
%
%    <CustomTools>
%      <Menu>
%       <Submenu name="Tracks Functions">
%        <Item name="Translate Tracks" icon="Matlab" tooltip="Translate tracks same starting point.">
%          <Command>MatlabXT::XTTrackTranslate(%i)</Command>
%        </Item>
%       </Submenu>
%      </Menu>
%      <SurpassTab>
%        <SurpassComponent name="bpSpots">
%          <Item name="Translate Tracks" icon="Matlab" tooltip="Translate tracks same starting point.">
%            <Command>MatlabXT::XTTrackTranslate(%i)</Command>
%          </Item>
%        </SurpassComponent>
%        <SurpassComponent name="bpSurfaces">
%          <Item name="Translate Tracks" icon="Matlab" tooltip="Translate tracks same starting point.">
%            <Command>MatlabXT::XTTrackTranslate(%i)</Command>
%          </Item>
%        </SurpassComponent>
%      </SurpassTab>
%    </CustomTools>
% 
%
%  Description:
%   
%   Copy the objects and translate them so that all tracks have the same 
%       starting point as the selected one.
%
%

function XTTrackTranslate(aImarisApplicationID)

% connect to Imaris interface
if ~isa(aImarisApplicationID, 'Imaris.IApplicationPrxHelper')
  javaaddpath ImarisLib.jar
  vImarisLib = ImarisLib;
  if ischar(aImarisApplicationID)
    aImarisApplicationID = round(str2double(aImarisApplicationID));
  end
  vImarisApplication = vImarisLib.GetApplication(aImarisApplicationID);
else
  vImarisApplication = aImarisApplicationID;
end

% the user has to create a scene with some tracks
vSurpassScene = vImarisApplication.GetSurpassScene;
if isequal(vSurpassScene, [])
  msgbox('Please create some tracks in the surpass scene!');
  return
end

% get the track
vObjectIndices = GetFirstSelectedTrack(vImarisApplication);
if isempty(vObjectIndices)  
  msgbox('Please select a track!');
  return
end

[vCoords, vTimes] = GetObjectsCoordsTimes( ...
  vImarisApplication, vObjectIndices);

% start point of selected track
[vValue, vIndexMin] = min(vTimes);
vStart = vCoords(vIndexMin, :);

vObjects = vImarisApplication.GetSurpassSelection;
vFactory = vImarisApplication.GetFactory;
if vFactory.IsSpots(vObjects)
  vObjects = vFactory.ToSpots(vObjects);
else
  vObjects = vFactory.ToSurfaces(vObjects);
end
vEdges = vObjects.GetTrackEdges;
vTracksIndices = SplitEdgesListIntoTracks(vEdges);
vTrackCount = numel(vTracksIndices);

vProgressDisplay = waitbar(0, 'Translating tracks');

if vFactory.IsSpots(vObjects)
  % spots
  vCoords = vObjects.GetPositionsXYZ;
  vTimes = vObjects.GetIndicesT;
  vRadii = vObjects.GetRadiiXYZ;
  for vIndex = 1:vTrackCount
    vIndices = vTracksIndices{vIndex} + 1;
    [vValue, vIndexMin] = min(vTimes(vIndices));
    vTranslation = vStart - vCoords(vIndices(vIndexMin), :);
    for vDim = 1:3
      vCoords(vIndices, vDim) = vCoords(vIndices, vDim) + vTranslation(vDim);
    end

    waitbar(vIndex/numel(vTrackCount), vProgressDisplay)
  end
  
  vResult = vFactory.CreateSpots;
  vResult.Set(vCoords, vTimes, vRadii(:,1));
  vResult.SetRadiiXYZ(vRadii);
else
  % surfaces
  vSize = vObjects.GetNumberOfSurfaces;
  vTimes = 1:vSize;
  for vIndex = 1:vSize
    vTimes(vIndex) = vObjects.GetTimeIndex(vIndex - 1);
  end  
  % compute translations
  vTrackLabels = 1:vSize;
  vTranslations = zeros(vTrackCount, 3);
  for vIndex = 1:vTrackCount
    vIndices = vTracksIndices{vIndex} + 1;
    [vValue, vIndexMin] = min(vTimes(vIndices));
    vCoords = vObjects.GetCenterOfMass(vIndices(vIndexMin) - 1);
    vTranslations(vIndex, :) = vStart - vCoords;
    vTrackLabels(vIndices) = vIndex;

    waitbar(vIndex/vTrackCount/2, vProgressDisplay)
  end
    vResult = vFactory.CreateSurfaces;
  for vIndex = 1:vSize
    vVertices = vObjects.GetVertices(vIndex - 1);
    vTriangles = vObjects.GetTriangles(vIndex - 1);
    vNormals = vObjects.GetNormals(vIndex - 1);
    
    vTranslation = vTranslations(vTrackLabels(vIndex), :);
    for vDim = 1:3
      vVertices(:, vDim) = vVertices(:, vDim) + vTranslation(vDim);
    end

    vResult.AddSurface(vVertices, vTriangles, vNormals, vTimes(vIndex));
    waitbar(0.5 + vIndex/vSize/2, vProgressDisplay)
  end
end

vResult.SetTrackEdges(vEdges);
vRGBA = vObjects.GetColorRGBA;
vResult.SetColorRGBA(vRGBA);
vSelection = vObjects.GetSelectedIndices;
vResult.SetSelectedIndices(vSelection);
vResult.SetName([char(vObjects.GetName), ' translated']);
vObjects.SetVisible(0);
vObjects.GetParent.AddChild(vResult, -1);

close(vProgressDisplay);


%---------------------------------------------------------%

function aObjectIndices = GetFirstSelectedTrack(aImarisApplication)

% aTracks is the list of indices (starting from 0) of the objects
%   in the track, *not* a set of edges
aObjectIndices = [];

% get the selected object (spots or surfaces)
vFactory = aImarisApplication.GetFactory;
vObject = vFactory.ToSpots(aImarisApplication.GetSurpassSelection);
if ~isempty(vObject)
  vSize = numel(vObject.GetIndicesT);
else
  vObject = vFactory.ToSurfaces(aImarisApplication.GetSurpassSelection);
  if ~isempty(vObject)
    vSize = vObject.GetNumberOfSurfaces;
  else
    return
  end
end

% get the sub-selection and the traks
vSelection = vObject.GetSelectedIndices;
vEdges = vObject.GetTrackEdges';
if isempty(vSelection) || isempty(vEdges)
  return
end

% get the first selected spot which belongs to a of the track
vCurrent = [];
vIndex = 0;
while ~any(vCurrent) && vIndex < numel(vSelection)
  vIndex = vIndex + 1;
  vCurrent = any(vEdges == vSelection(vIndex));
end
if ~any(vCurrent(:))
  % no spots selected belongs to a track
  return
end

% get the spots of the track of the first selected object
vTrackBool = false(vSize, 1);
vTrackBool(vSelection(vIndex) + 1) = true;
while any(vCurrent)
  vObjects = vEdges([vCurrent; vCurrent]);
  vTrackBool(vObjects + 1) = true;
  vEdges([vCurrent; vCurrent]) = vSize;
  vCurrent = false(size(vCurrent));
  for vIndex = 1:numel(vObjects)
    vCurrent = vCurrent | any(vEdges == vObjects(vIndex));
  end
end
vIndices = 0:vSize-1;
aObjectIndices = vIndices(vTrackBool);


%---------------------------------------------------------%

function [aCoords, aTimes] = GetObjectsCoordsTimes( ...
  aImarisApplication, aSelectionIndices)

aCoords = [];
aTimes = [];

% get the selected object (spots or surfaces)
vFactory = aImarisApplication.GetFactory;
vSpots = vFactory.ToSpots(aImarisApplication.GetSurpassSelection);
vSurfaces = vFactory.ToSurfaces(aImarisApplication.GetSurpassSelection);

if ~isempty(vSpots)
  aCoords = vSpots.GetPositionsXYZ;
  aTimes = vSpots.GetIndicesT;
  vSize = numel(aTimes);
  % remove invalid selection indices
  aSelectionIndices = aSelectionIndices(aSelectionIndices < vSize) + 1;
  % perform selection
  aCoords = aCoords(aSelectionIndices, :);
  aTimes = aTimes(aSelectionIndices);
elseif ~isempty(vSurfaces)
  vSize = vSurfaces.GetNumberOfSurfaces;
  % remove invalid selection indices
  aSelectionIndices = aSelectionIndices(aSelectionIndices < vSize);
  vSize = numel(aSelectionIndices);
  aCoords = zeros(vSize, 3);
  aTimes = zeros(vSize, 1);
  % read selection
  for vIndex = 1:vSize
    vSurface = aSelectionIndices(vIndex);
    aCoords(vIndex, :) = vSurfaces.GetCenterOfMass(vSurface);
    aTimes(vIndex) = vSurfaces.GetTimeIndex(vSurface);
  end
end


%---------------------------------------------------------%

function aTracksIndices = SplitEdgesListIntoTracks(aEdges)

% compute the labels (one label for each connected set of objects)
vNumberOfObjects = max(aEdges(:)) + 1;
vLabels = zeros(vNumberOfObjects, 1);
vEdgeLabels = 1:size(aEdges, 1);

for vIndex = 1:size(aEdges, 1)
  vEdge = aEdges(vIndex, :) + 1;
  vLabel = vLabels(vEdge);
  vLabel = vLabel(vLabel > 0);
  
  if isempty(vLabel) 
    % no label yet -> create new
    vLabel = max(vLabels) + 1;
  elseif numel(vLabel) > 1 
    % multiple labels assigned to connected objects -> merge
    vFirst = min(vLabel);
    for vLabIndex = 1:numel(vLabel)
      vLab = vLabel(vLabIndex);
      if vLab ~= vFirst
        vLabels(vLabels == vLab) = vFirst;
        vLabels(vLabels > vLab) = vLabels(vLabels > vLab) - 1;
        vEdgeLabels(vEdgeLabels == vLab) = vFirst;
        vEdgeLabels(vEdgeLabels > vLab) = vEdgeLabels(vEdgeLabels > vLab) - 1;
      end
    end
    vLabel = vFirst;
  end
  vLabels(vEdge) = vLabel;
  vEdgeLabels(vIndex) = vLabel;
end

vNumTracks = max(vLabels);
aTracksIndices = cell(vNumTracks, 1);
vIndices = 0:vNumberOfObjects-1;
for vIndex = 1:vNumTracks
  aTracksIndices{vIndex} = vIndices(vLabels == vIndex);
end
